



def plot_sensor_signals(df, sensors_signals_col, target_col_name):
        
    facetgrid = sns.FacetGrid(df, hue=target_col_name, height=5,aspect=3)
    facetgrid.map(sns.distplot, sensors_signals_col, hist=False).add_legend()


def signals_plot_by_target(df, target_col_name, labels):
    fig, axes = plt.subplots(nrows=24,ncols=2, figsize=(20,36))
    df[df[target_col_name]==labels[0]].iloc[:, 1:25].plot(ax=axes[:,0], subplots=True, sharex=True, figsize=(10,25))

    df[df[target_col_name]==labels[1]].iloc[:, 1:25].plot(ax=axes[:,1], subplots=True, sharex=True, figsize=(10,25));

    # plt.legend()

    axes[0][0].set_title(labels[0])
    axes[0][1].set_title(labels[1])

    plt.show();




def plot_by_exp_stage():
    fig, axes = plt.subplots(nrows=24,ncols=5, figsize=(20,36))
    df[df['exp_stage']=='baseline'].iloc[:, 1:25].plot(ax=axes[:,0], subplots=True, sharex=True, figsize=(10,25))

    df[df['exp_stage']=='absorb'].iloc[:, 1:25].plot(ax=axes[:,1], subplots=True, sharex=True, figsize=(10,25));
    df[df['exp_stage']=='pause'].iloc[:, 1:25].plot(ax=axes[:,2], subplots=True, sharex=True, figsize=(10,25));
    df[df['exp_stage']=='desorb'].iloc[:, 1:25].plot(ax=axes[:,3], subplots=True, sharex=True, figsize=(10,25));
    df[df['exp_stage']=='flush'].iloc[:, 1:25].plot(ax=axes[:,4], subplots=True, sharex=True, figsize=(10,25));

    # plt.legend()

    axes[0][0].set_title('baseline')
    axes[0][1].set_title('absorb')
    axes[0][2].set_title('pause')
    axes[0][3].set_title('desorb')
    axes[0][4].set_title('flush')





def plot_new_sample(model, x_train, x):
    fig, ax = plt.subplots()
    x = x.reshape(1,-1)
    ax.scatter(x_train[:,0], x_train[:,1], c=l)
    ax.scatter(x[:,0], x[:,1])
# 
def sensor_channel_response_bar(df,date_time,x_columns):
  df = df[x_columns]
  x = df.columns.tolist()
  # 
  fig = plt.figure()
  plt.title(str(date_time))
  for i in range(len(df)):
    y = df.iloc[i].values.tolist()
    plt.bar(x, y)
    plt.show()
    plt.ylim([0,500])
    plt.ylabel('mV')
    fig = plt.gcf()
    fig.autofmt_xdate()
    sleep(0.05)
    clear_output(wait=True)

def sensor_channel_response_line(df,date_time,x_columns):
  df = df[x_columns]
  x = df.columns.tolist()
  y = df.values.tolist()
  # 
  fig = plt.figure()
  plt.title(str(date_time))
  plt.plot(y)
  plt.ylim([0,500])
  plt.ylabel('mV')
  fig = plt.gcf()
  fig.autofmt_xdate()
  plt.legend(x)
  plt.show()




def plot_voc_signal_sensors(control, covid):

    fig, axes = plt.subplots(nrows=24,ncols=1, figsize=(25,42))

    control.iloc[:, 3:27].plot(subplots=True, sharex=True, figsize=(10,25), color='b', label='Control', legend=False, ax=axes);
    covid.iloc[:, 3:27].plot(subplots=True, sharex=True, figsize=(10,25), color='r', label='Covid', legend=False, ax=axes);
    axes[0].set_title('VOC Signature After Normalisation', fontsize=20);
    fig.text(0.5, 0.11, 'Time Steps', ha='center', fontsize=16);
    fig.text(0.1, 0.5, 'Magnitude of Voltage Signals', va='center', rotation='vertical', fontsize=16);
    plt.legend({'Control','Covid'}, fontsize=16);
    fig.savefig('new_data.jpeg');


def plot_voc_sensor(covid, control):

    fig, ax = plt.subplots(figsize=(14, 6), dpi=80)
    ax.plot(covid['sensor_1'], label='Covid', color='blue', animated = True, linewidth=1)
    ax.plot(control['sensor_1'], label='Control', color='red', animated = True, linewidth=1)


def plot_sensor_mean(df, cols):
    ax = df[cols].mean().plot(kind='bar', figsize=(10,6), title="Features Standard Deviation", label='covid')
    df[cols].mean().plot(kind='bar', figsize=(10,6), title="Features Standard Deviation", ax=ax, color='red', alpha=0.3, label='control');
    plt.legend();


import boto3
import os
from dotenv import load_dotenv
load_dotenv()
from decouple import config


#  AWS_ACCESS_KEY = os.environ['AWS_ACCESS_KEY']
#  AWS_SECRET_KEY = os.environ['AWS_SECRET_KEY']

AWS_ACCESS_KEY = config('AWS_ACCESS_KEY')
AWS_SECRET_KEY = config('AWS_SECRET_KEY')


def upload_file(file_name, bucket):
    """
    Function to upload a file to an S3 bucket
    """
    object_name = file_name
    s3_client = boto3.client('s3', aws_access_key_id=AWS_ACCESS_KEY, aws_secret_access_key=AWS_SECRET_KEY)
    response = s3_client.upload_file(file_name, bucket, object_name)

    return response




def download_file(file_name, bucket):
    """
    Function to download a given file from an S3 bucket
    """
    download_folder = 'downloads/'
    if not os.path.exists(os.path.join(os.getcwd(), download_folder)):
        os.mkdir(os.path.join(os.getcwd(), download_folder))

    s3_client = boto3.client('s3', aws_access_key_id=AWS_ACCESS_KEY, aws_secret_access_key=AWS_SECRET_KEY)

    response = s3_client.download_file(bucket, file_name, f'{download_folder}/data.csv')

    return response




def list_files(bucket):
    """
    Function to list files in a given S3 bucket
    """
    s3 = boto3.client('s3')
    contents = []
    for item in s3.list_objects(Bucket=bucket)['Contents']:
        contents.append(item)

    return contents



from sklearn.metrics import accuracy_score, recall_score, precision_score
import matplotlib.pyplot as plt



def eval_metrics(actual, pred):
    """evaluates model performance

    Args:
        actual (numpy array): true labels
        pred (numpy array): predicited labels

    Returns:
        tuple: returns a tuple of accuracy, sensitivity, specificity, precision
    """
    
    accuracy = accuracy_score(actual, pred)
    sensitivity = recall_score(actual, pred)
    specificity = recall_score(actual, pred, pos_label=0)
    precision = precision_score(actual, pred)

    return accuracy, sensitivity, specificity, precision






def make_plots():
  plot_confusion_matrix(clf, X_test, y_test)  
  skplt.metrics.plot_roc_curve(y_true, y_probas)
  plt.show()




cmap = plt.cm.Blues

def confusion_matrix_plot(model,X_test,X_train, y_train, y_test):
  # 
  y_pred = model.predict(X_test)
  print("Accuracy: %.2f%%" % (accuracy_score(y_test, y_pred) * 100))
  print("Recall: %.2f%%" % (recall_score(y_test, y_pred, average="binary", pos_label='Covid') * 100.0))
  print("Precision: %.2f%%" % (precision_score(y_test, y_pred, average="binary", pos_label='Covid') * 100.0))

  plot_confusion_matrix(model, X_test, y_test,
                            display_labels=['Covid', 'Non-Covid'], cmap=cmap)
  plt.savefig('confusion_matrix.eps')

  fea_imp = pd.DataFrame({'imp': model.feature_importances_, 'col': X_train.columns})
  fea_imp = fea_imp.sort_values(['imp', 'col'], ascending=[True, False]).iloc[-30:]
  fea_imp.plot(kind='barh', x='col', y='imp', figsize=(10, 7), legend=None)
  plt.title('XGBoost - Feature Importance')
  plt.ylabel('Features')
  plt.xlabel('Importance')
  plt.savefig('feature_importance.eps')

  disp = plot_precision_recall_curve(model, X_test, y_test)
  disp.ax_.set_title('Binary class Precision-Recall curve')
  plt.savefig('precision_recall_curve.eps')




def model_explanation(model,X,X_test):
  # 
  explainer = shap.TreeExplainer(model)
  shap_values = explainer.shap_values(X_test) 
  # shap.initjs()
  # 
  # Summary plot
  shap.summary_plot(shap_values, X_test)
  # 
  for name in X.columns:
    shap.dependence_plot(name, shap_values, X_test, display_features = X)


def random_prediction(X_test,y_test,model):
  explainer = shap.TreeExplainer(model)
  shap_values = explainer(X_test)
  i = random.randint(0, len(X_test))
  shap.plots.waterfall(shap_values[i])
  print('Category: ' + str(y_test.values[i][0]))
  print('Prediction: ' + str(model.predict(X_test.iloc[[i]])[0]))
  print('Prediction Probability: ' + str(model.predict_proba(X_test.iloc[[i]])[0]))









def calculate_prediction_and_probability(list_files, df_list, x_columns, model):
  # 
  prediction = []
  probability = []
  for i in range(len(list_files)):
    df_temp = df_list[i]
    df_temp = df_temp[x_columns]
    prediction.append([])
    probability.append([])
    for j in range(df_temp.count()[0]):
      input_data_sample = df_temp.iloc[[j]]
      prediction_temp =  int(model.predict(input_data_sample))
      prediction[i].append(prediction_temp)
      probability_temp =  model.predict_proba(input_data_sample)
      probability_temp = float(probability_temp[0][1])
      probability[i].append(probability_temp)
  
  return prediction, probability

def calculate_mean_probability(list_files, probability):
  # 
  mean_probability = []
  std_probability = []
  for i in range(len(list_files)):
    probability_array = np.array(probability[i])
    # 
    mean_probability_temp = np.mean(probability_array)
    mean_probability.append(mean_probability_temp)
    # 
    std_probability_temp = np.std(probability_array)
    std_probability.append(std_probability_temp)
  # 
  return mean_probability, std_probability

def plot_histogram_probabilities(mean_probability):
  # 
  plt.hist(mean_probability, 30) 
  plt.title("histogram") 
  plt.show()

def plot_probability_over_time(date_time_list,mean_probability):
  # 
  plt.style.use('seaborn-colorblind')
  plt.scatter(date_time_list,mean_probability)
  degrees = 70
  plt.axhline(y=0.5, color='r', linestyle='--', linewidth = '1')
  plt.xticks(rotation=degrees)
  plt.xlabel('Time')
  plt.ylabel('Probability of Covid')
  plt.ylim([0,1])
  plt.title('Probability of COVID vs. Time')
  time_now = str(datetime.now())
  plt.savefig('Plot'+time_now +'.png', dpi = 200, bbox_inches='tight')






def plot_algorithm_performance(names, results, metrics, savefig=False, ylim=False):

    fig = plt.figure(figsize=(16,6))
    fig.suptitle('Comparison of algorithm performance')
    ax = fig.add_subplot(111)
    plt.boxplot(results)
    ax.set_xticklabels(names)
    plt.ylabel(metrics)
    plt.xlabel('models')

    if savefig:
        fig.savefig(f'{metrics}.jpeg', transparent = False)
    
    if ylim:
        plt.ylim([0,1])
        
    plt.show()



def run_models_cv(models, X_train, y_train, scoring='accuracy'):
    results = []
    names = []

    if scoring == 'specificity':
        scoring = make_scorer(recall_score, pos_label=0)
    else:
        for name, model in models:
            kfold = model_selection.KFold(n_splits=10, random_state=seed, shuffle=True)
            cv_results = model_selection.cross_val_score(model, X_train, y_train, cv=kfold, scoring=scoring)
            results.append(cv_results)
            names.append(name)
      
    return names, results, scoring






import pandas as pd
import glob
import os
from datetime import datetime
import pickle
import random
import matplotlib.pyplot as plt
import numpy as np
import re
from datetime import timedelta




SENSOR_FEATURES = ['sensor_1','sensor_2','sensor_3','sensor_4','sensor_5',
 'sensor_6', 'sensor_7', 'sensor_8', 'sensor_9', 'sensor_10', 'sensor_11', 'sensor_12', 'sensor_13',
 'sensor_14', 'sensor_15', 'sensor_16', 'sensor_17', 'sensor_18', 'sensor_19', 'sensor_20', 'sensor_21', 'sensor_22', 'sensor_23', 'sensor_24']



def find_header_value(field, filename):
  
  
    # Get first 6 lines
    with open(filename) as f:
        for line_no, line in enumerate(f): 
            if line.startswith(field):
                # Get the part of the string after the field name
                end_of_string = line[len(field):]
                string = end_of_string[:-1]
                break
        return string, line_no


def get_exp_stage(time, data_points_per_sec=4, baseline=2, absorb=7, pause=1, desorb=5, flush=23):
    
    # data_points_per_sec: data point per second

    baseline_time = baseline*data_points_per_sec
    if time <= baseline_time:
        return 'baseline'
    absorb_time = baseline_time + absorb*data_points_per_sec
    if time <= absorb_time:
        return 'absorb'
    pause_time = absorb_time + pause*data_points_per_sec
    if time <= pause_time:
        return 'pause'
    desorb_time = pause_time + desorb*data_points_per_sec
    if time <= desorb_time:
        return 'desorb'
    flush_time = desorb_time + flush*data_points_per_sec
    if time <= flush_time:
        return 'flush'
    wait_time = flush_time + flush*data_points_per_sec
    if time <= wait_time:
        return 'wait'


def rename_columns(df):

    for col in df.columns:
        if 'Sen' in col:
            new_col = re.findall('\d+', col)[0]
            df.rename(columns={col:f'sensor_{new_col}'}, inplace=True)
    df.rename(columns={'Data Points': 'timesteps'}, inplace=True)
    df.rename(columns={'Humidity (%r.h.)':'humidity'}, inplace=True)

    return df







def get_label(f):   

    filename = f.split('/')[-1]
    exp_name, _ = find_header_value('Name of the experiment =', f)

    if ("Neg" in filename) or ("Control" in filename) or ('CONTROL' in filename) or ("Neg" in exp_name) or ("Control" in exp_name) or ('CONTROL' in exp_name) or ("Neg" in filename):
        return "Control"
    elif ("Pos" in filename) or ("Covid" in filename) or ('COVID' in filename) or ("Pos" in exp_name) or ("Covid" in exp_name) or ('COVID' in exp_name) or ("Pos" in filename):
        return "Covid"
    else:
        return "Other"

# def get_label(f):

#     filename = f.split('/')[-1]
   

#     exp_name, _ = find_header_value('Name of the experiment =', f)

#     if ("Neg" in exp_name) or ("Control" in exp_name) or ('CONTROL' in exp_name) or ("Neg" in filename):
#         return "Control"
#     elif ("Pos" in exp_name) or ("Covid" in exp_name) or ('COVID' in exp_name) or ("Pos" in filename):
#         return "Covid"
#     elif (exp_name[-2].islower()):
#         return "Control"
#     else:
#         return "Other"
    

def get_exp_stage_duration(f):

    baseline, _ = find_header_value('BaseLine = ', f)
    absorb, _ = find_header_value('Absorb = ', f)
    pause, _ = find_header_value('Pause = ', f)
    desorb, _ = find_header_value('Desorb = ', f)
    flush, _ = find_header_value('Flush = ', f)

    return float(baseline), float(absorb), float(pause), float(desorb), float(flush)



def offset_one_sample(df_temp, sec=2):

    datapoints = sec*4
    baseline = df_temp[SENSOR_FEATURES].iloc[:datapoints].mean().values

    df_temp[SENSOR_FEATURES] = df_temp[SENSOR_FEATURES].apply(lambda row: row - baseline, axis=1)

    return df_temp


def offset_batch_samples(df, sec=2, id_col='exp_unique_id'):
    """normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods 

    Args:
        df (pandas dataframe): dataframe of cyclic sensor data
        sec (int, optional): number of seconds to use for offsetting. Defaults to 2.
        id_col (str, optional): unique id of each cycle of experiment. Defaults to 'exp_unique_id'.

    Returns:
        pandas dataframe: _description_
    """
    samples = df.groupby(id_col)
    df_list = []
    for i, sample in samples:
        df_list.append(offset_one_sample(sample, sec=sec))
    return pd.concat(df_list)



def fillna_columns(df_temp):

    num = df_temp._get_numeric_data()
    num[num < 0] = np.nan
    num = num.fillna(num.mean())
    return df_temp


def preprocess_single_file(f, parse_time=True, parse_filename=True, rename_column=True, fillna=True, offset=True, has_label=True):
    """preprocess a single cycle (one txt file) of experiment

    Args:
        f (str): file path
        parse_time (bool, optional): extract time features. Defaults to True.
        parse_filename (bool, optional): adds filename to dataframe. Defaults to True.
        rename_column (bool, optional): rename columns. Defaults to True.
        fillna (bool, optional): replaces extreme negative values with mean of column. Defaults to True.
        offset (bool, optional): normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods . Defaults to True.
        has_label (bool, optional): extracts label and add to dataframe. Defaults to True.

    Returns:
        pandas dataframe: dataframe of features in a cycle(one txt file)
    """   

    _, line_no = find_header_value('Data Points', f)

    # print(line_no)

    df_temp = pd.read_csv(f,sep='\t',header=(line_no - 1))
    baseline, absorb, pause, desorb, flush = get_exp_stage_duration(f)
    # print(df_temp.head())
    df_temp['measurement_stage'] = df_temp['Data Points'].apply(get_exp_stage)
    
    if parse_time:
      date, _ = find_header_value('Date = ', f)
      df_temp['date_exp'], _ = find_header_value('Date = ', f)
      df_temp['time_elapsed'] = df_temp.index / 4
      time_start, _ = find_header_value('Time = ', f)
      time_elapsed = df_temp.index / 4
      timestamp = pd.to_datetime(date + " " + time_start)
      df_temp['datetime_exp_start'] = time_start
      df_temp['datetime_exp'] = pd.to_datetime(date + " " + time_start) + pd.to_timedelta(time_elapsed, unit='s')

    if parse_filename:
      df_temp['filename'] = f.split('/')[-1]
    
    if has_label:
        df_temp['result'] = get_label(f)
    
    df_temp['exp_name'] = find_header_value('Name of the experiment = ', f)[0][1:-1]

    

    if rename_column:
      df = rename_columns(df_temp)


    if fillna:
        df_temp = fillna_columns(df_temp)


    if offset:
        df_temp = offset_one_sample(df_temp)


    return df_temp

def get_all_files(parent_dir):
    all_files = []
    for path, subdirs, files in os.walk(parent_dir):
        for name in files:
            if (name.endswith('txt')) and (len(name) > 25) and ('Wash' not in name):
                all_files.append(os.path.join(path, name))
    return all_files

def get_unique_files(df):
    """select one (chosen at random) of repeat experiemnts

    Args:
        df (pandas dataframe): dataframe of cyclic sensor data with repeat experiments

    Returns:
        pandas dataframe: dataframe with only one of each repeat experiemnt
    """

    g = df.groupby(['exp_name'])

    exp_names = list(set(df['exp_name']))
    all = []
    for i in exp_names:
        df = g.get_group(i)

        uniq_files = list(df['filename'].unique())
        file_chosen = random.choice(uniq_files)
        df = df[df['filename'] == file_chosen]
        all.append(df)
    return pd.concat(all)


def reorder_columns(df):

  col_list = df.columns.tolist()
  new_col_list = [col_list[-1]] + [col_list[-2]]  + col_list[:-2] 
  df = df[new_col_list]

  return df

def preprocess_all_files(path_to_measurements, parse_time=True, parse_filename=True, concat_files=True, rename_column=True, no_repeat=False, fillna=True, offset=False, has_label=True):
    """preprocess all cycle (all txt file) of experiment

    Args:
        path_to_measurements (str): folder path to measurements
        parse_time (bool, optional): extract time features. Defaults to True.
        parse_filename (bool, optional): adds filename to dataframe. Defaults to True.
        rename_column (bool, optional): rename columns. Defaults to True.
        no_repeat (bool, optional): _description_. Defaults to False.
        fillna (bool, optional): replaces extreme negative values with mean of column. Defaults to True.
        offset (bool, optional): normalises all sensors in each cycle of experiment by subtracting the average of the sensor voltage in the 1st 2 sec from sensor voltage in other periods . Defaults to True.
        has_label (bool, optional): extracts label and add to dataframe. Defaults to True.

    Returns:
        pandas dataframe: dataframe of features in a cycle(one txt file)
    """


    # if path_to_measurements:
    #     list_files = glob.glob(os.path.join(path_to_measurements,"*.txt"))
    # else:
    #     list_files = file_list

    list_files = get_all_files(path_to_measurements)
    # print('len of files: ', len(list_files))
    # print(list_files)

    df_list = []
    for i, f in enumerate(list_files):

        df_temp = preprocess_single_file(f, parse_time=parse_time, parse_filename=parse_filename, offset=offset, fillna=fillna, has_label=has_label)
        df_temp['exp_unique_id'] = i
        df_list.append(df_temp)

    
    df = pd.concat(df_list)
    df = reorder_columns(df)


    if no_repeat:
        return get_unique_files(df)
    else:
        return df








  